﻿using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Storage;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Media;
using Microsoft.Xna.Framework.Input.Touch;
using Windows.Devices.Sensors;
using Microsoft.Xna.Framework.Audio;


namespace MonoOno
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class SmudgeGame : Game
    {
        GraphicsDeviceManager _graphics;
        SpriteBatch _spriteBatch;

        #region Sprites

        interface ISprite
        {
            void Reset(SamplerState game);
            void Draw(SmudgeGame game);
            void Update(SmudgeGame game);
        }

        class Sprite : ISprite
        {
            protected Texture2D spriteTexture;
            protected Vector2 spritePosition;
            protected float spriteScale;
            protected float spriteRotation;
            protected Color spriteColor;
            protected Vector2 spriteOrigin;

            public void Reset(SamplerState game)
            {
            }

            public virtual void Draw(SmudgeGame game)
            {
                game._spriteBatch.Draw(
                    spriteTexture,  // texture of sprite
                    spritePosition, // position on screen
                    null,           // source rectangle (all of it)
                    spriteColor,    // color of the light
                    spriteRotation, // rotation
                    spriteOrigin,   // centre of sprite
                    spriteScale,    // scale
                    SpriteEffects.None,
                    1);
            }

            public Sprite(SmudgeGame game, Texture2D texture, Vector2 position, Color color, float widthFactor, float rotation)
            {
                spriteTexture = texture;
                spritePosition = position;
                spriteColor = color;
                spriteRotation = rotation;
                spriteOrigin = new Vector2(texture.Width / 2, texture.Height / 2);

                float spriteWidth = game.GraphicsDevice.Viewport.Width / widthFactor;

                spriteScale = spriteWidth / texture.Width;
            }

            public virtual void Update(SmudgeGame game)
            {
            }
        }

        class DriftingSprite : Sprite, ISprite
        {
            protected Vector2 spriteSpeed;
            protected float spriteRotationSpeed;

            public DriftingSprite(SmudgeGame game, Texture2D texture, Vector2 position, Color color, float widthFactor, float rotation, Vector2 speed, float rotationSpeed) :
                base(game, texture, position, color, widthFactor, rotation)
            {
                spriteSpeed = speed;
                spriteRotationSpeed = rotationSpeed;
            }

            public override void Update(SmudgeGame game)
            {
                base.Update(game);
                spritePosition += spriteSpeed;
                spriteRotation += spriteRotationSpeed;

                float halfWidth = (spriteTexture.Width * spriteScale) / 2;

                if (spriteSpeed.X > 0)
                {
                    // moving left to right - check right edge
                    float leftEdge = spritePosition.X - halfWidth;
                    if (leftEdge > game.GraphicsDevice.Viewport.Width)
                        spritePosition.X = -halfWidth;
                }
                else
                {
                    // moving right to left - check left edge
                    float rightEdge = spritePosition.X + halfWidth;
                    if (rightEdge < 0)
                        spritePosition.X = game.GraphicsDevice.Viewport.Width + halfWidth;
                }

                float halfHeight = (spriteTexture.Height * spriteScale) / 2;

                if (spriteSpeed.Y > 0)
                {
                    // moving down the screen - check off the bottom
                    float bottomEdge = spritePosition.Y - halfHeight;
                    if (bottomEdge > game.GraphicsDevice.Viewport.Height)
                        spritePosition.Y = -halfHeight;
                }
                else
                {
                    // moving up the screen - check off the top
                    float topEdge = spritePosition.Y + halfHeight;
                    if (topEdge < 0)
                        spritePosition.Y = game.GraphicsDevice.Viewport.Height + halfHeight;
                }
            }
        }

        class BouncingSprite : Sprite, ISprite
        {
            protected Vector2 spriteSpeed;
            protected float spriteRotationSpeed;

            public BouncingSprite(SmudgeGame game, Texture2D texture, Vector2 position, Color color, float widthFactor, float rotation, Vector2 speed, float rotationSpeed) :
                base(game, texture, position, color, widthFactor, rotation)
            {
                spriteSpeed = speed;
                spriteRotationSpeed = rotationSpeed;
            }

            public override void Update(SmudgeGame game)
            {
                base.Update(game);
                spritePosition += spriteSpeed;
                spriteRotation += spriteRotationSpeed;

                float halfWidth = (spriteTexture.Width * spriteScale) / 2;

                float leftEdge = spritePosition.X - halfWidth;
                if (leftEdge < 0)
                    spriteSpeed.X = Math.Abs(spriteSpeed.X);

                float rightEdge = spritePosition.X + halfWidth;
                if (rightEdge > game.GraphicsDevice.Viewport.Width)
                    spriteSpeed.X = -Math.Abs(spriteSpeed.X);

                float halfHeight = (spriteTexture.Height * spriteScale) / 2;

                float bottomEdge = spritePosition.Y - halfHeight;
                if (bottomEdge < 0)
                    spriteSpeed.Y = Math.Abs(spriteSpeed.Y);

                float topEdge = spritePosition.Y + halfHeight;
                if (topEdge > game.GraphicsDevice.Viewport.Height + halfHeight)
                    spriteSpeed.Y = -Math.Abs(spriteSpeed.Y);
            }
        }

        class FallingBounceSprite : BouncingSprite, ISprite
        {
            protected float spriteMass;
            protected float spriteFriction;

            public FallingBounceSprite(SmudgeGame game, Texture2D texture, Vector2 position, Color color, float widthFactor, float rotation, Vector2 speed, float rotationSpeed, float mass, float friction) :
                base(game, texture, position, color, widthFactor, rotation, speed, rotationSpeed)
            {
                spriteMass = mass;
                spriteFriction = friction;
            }

            public override void Update(SmudgeGame game)
            {
                Vector3 weightFactor = game.Acceleration * spriteMass;
                spriteSpeed.X += weightFactor.X;
                spriteSpeed.Y -= weightFactor.Y;
                spriteSpeed = spriteSpeed * spriteFriction;
                base.Update(game);
            }

        }


        class FallingDriftingSprite : DriftingSprite, ISprite
        {
            protected float spriteMass;
            protected float spriteFriction;

            public FallingDriftingSprite(SmudgeGame game, Texture2D texture, Vector2 position, Color color, float widthFactor, float rotation, Vector2 speed, float rotationSpeed, float mass, float friction) :
                base(game, texture, position, color, widthFactor, rotation, speed, rotationSpeed)
            {
                spriteMass = mass;
                spriteFriction = friction;
            }

            public override void Update(SmudgeGame game)
            {
                Vector3 weightFactor = game.Acceleration * spriteMass;
                spriteSpeed.X += weightFactor.X;
                spriteSpeed.Y -= weightFactor.Y;
                spriteSpeed = spriteSpeed * spriteFriction;
                base.Update(game);
            }

        }

        class ChasingDriftingSprite : DriftingSprite
        {
            protected float spriteAttraction;
            protected float spriteFriction;

            public ChasingDriftingSprite(SmudgeGame game, Texture2D texture, Vector2 position, Color color, float widthFactor, float rotation, Vector2 speed, float rotationSpeed, float attraction, float friction) :
                base(game, texture, position, color, widthFactor, rotation, speed, rotationSpeed)
            {
                spriteAttraction = attraction;
                spriteFriction = friction;
            }

            public override void Update(SmudgeGame game)
            {
                base.Update(game);
            }


        }

        class GrowingSprite : Sprite, ISprite
        {
            public GrowingSprite(SmudgeGame game, Texture2D texture, Vector2 position, Color color, float widthFactor, float rotation) :
                base(game, texture, position, color, widthFactor, rotation)
            {
            }


            float spriteGrowSpeed = 0.01f;
            int spriteGrowLimit = 10;
            int spriteGrowCount = 0;

            public override void Update(SmudgeGame game)
            {
                if (spriteGrowCount == spriteGrowLimit) return;

                spriteGrowCount++;

                spriteScale += spriteGrowSpeed;
                base.Update(game);
            }
        }

        #endregion

        #region Colours

        System.Random colorRand = new Random();

        Color randomColor()
        {
            int r = colorRand.Next(256);
            int g = colorRand.Next(256);
            int b = colorRand.Next(256);
            return Color.FromNonPremultiplied(r, g, b, 255);
        }

        #endregion

        #region Accelerometer

        public Vector3 Acceleration;

        Accelerometer accel;

        void startAccelerometer()
        {
            accel = Accelerometer.GetDefault();

            accel.ReadingChanged += accel_ReadingChanged;
        }

        void accel_ReadingChanged(Accelerometer sender, AccelerometerReadingChangedEventArgs args)
        {
            Acceleration.X = (float)args.Reading.AccelerationX;
            Acceleration.Y = (float)args.Reading.AccelerationY;
            Acceleration.Z = (float)args.Reading.AccelerationZ;
        }

        #endregion

        #region Touch Panel

        public bool GotTouch = false;

        public Vector2 TouchPosition;

        #endregion

        // Game World

        Texture2D demoTexture;
        Rectangle demoRectangle;

        List<ISprite> gameSprites = new List<ISprite>();

        Texture2D smudgeTexture;

        SpriteFont font;

        SoundEffect dingSound;

        int smudgeLength = 1000;

        public SmudgeGame()
        {
            _graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            TouchPanel.EnabledGestures = GestureType.Tap | GestureType.FreeDrag | GestureType.DoubleTap ;

            startAccelerometer();

            base.Initialize();
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            _spriteBatch = new SpriteBatch(GraphicsDevice);

            demoTexture = Content.Load<Texture2D>("DemoArtwork");

            demoRectangle = new Rectangle(0, 0, GraphicsDevice.Viewport.Width, GraphicsDevice.Viewport.Height);

            smudgeTexture = Content.Load<Texture2D>("Smudge");

            font = Content.Load<SpriteFont>("MessageFont");

            dingSound = Content.Load<SoundEffect>("ding");
        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
            // TODO: Unload any non ContentManager content here
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            if (Keyboard.GetState().IsKeyDown(Keys.Escape))
                Exit();

            GotTouch = false;

            while (TouchPanel.IsGestureAvailable)
            {
                GestureSample gesture = TouchPanel.ReadGesture();

                switch (gesture.GestureType)
                {
                    case GestureType.DoubleTap:
                        gameSprites.Clear();

                        SoundEffectInstance dingInstance = dingSound.CreateInstance();

                        dingInstance.Play();

                        break;

                    case GestureType.Tap:
                    case GestureType.FreeDrag:
                        if (gameSprites.Count > smudgeLength)
                            gameSprites.RemoveAt(0);

                        gameSprites.Add(new GrowingSprite(this, smudgeTexture, gesture.Position, randomColor(), 20, 0));
                        ISprite movingsprite = new FallingDriftingSprite(
                            this,
                            smudgeTexture,
                            gesture.Position,
                            randomColor(),
                            20,
                            0,
                            Vector2.Zero,//new Vector2 (rand.Next(-5, 5), rand.Next(-5, 5)),
                            0,
                            .5f,
                            0.9f);
                        gameSprites.Add(movingsprite);

                        GotTouch = true;
                        TouchPosition = gesture.Position;

                        break;
                }
            }

            foreach (ISprite sprite in gameSprites)
                sprite.Update(this);

            base.Update(gameTime);
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.CornflowerBlue);

            _spriteBatch.Begin();

            _spriteBatch.Draw(demoTexture, demoRectangle, Color.White);

            _spriteBatch.DrawString(font, "Smudges : " + gameSprites.Count, Vector2.Zero, Color.White);

            foreach (ISprite sprite in gameSprites)
                sprite.Draw(this);

            _spriteBatch.End();

            base.Draw(gameTime);
        }
    }
}
